import json
import os

from sqlalchemy.future.engine import create_engine

from ata_db_models.helpers import (
    Component,
    DatabaseConnectionCredentials,
    Grant,
    Privilege,
    Stage,
    create_user,
    get_conn_string,
    get_ssm_client,
    grant_privileges,
)

if __name__ == "__main__":
    stages = [Stage.dev, Stage.prod]
    reader = Component(
        name="reader", grants=[Grant(privileges=[Privilege.SELECT], tables=["event", "prescription"])], policies=[]
    )

    for stage in stages:
        engine = create_engine(get_conn_string())
        ssm_client = get_ssm_client()
        username = f"{stage}_{reader.name}"
        with engine.connect() as conn:
            pw = create_user(conn, username=username, bypassrls=True)
            conn.commit()

        stage_engine = create_engine(get_conn_string(db_name=stage))
        with stage_engine.connect() as conn:
            for grant in reader.grants:
                for table in grant.tables:
                    grant_privileges(conn=conn, user_or_role=username, table=table, privileges=grant.privileges)
            conn.commit()

        if ssm_client:
            user_creds = DatabaseConnectionCredentials(
                HOST=os.getenv("HOST", "localhost"),
                PORT=int(os.getenv("PORT", "5432")),
                DB_NAME=stage,
                USERNAME=username,
                PASSWORD=pw,
            )
            ssm_client.put_parameter(
                Name=f"/{stage}/ata/{reader.name}/all/database-credentials",
                Description=f"DB credentials for a read-only user, for {stage}.",
                Value=json.dumps(user_creds.__dict__),
                Type="String",
                Overwrite=True,
            )



from ata_db_models.helpers import (
    Component,
    Grant,
    Partner,
    Privilege,
    RowLevelSecurityPolicy,
    Stage,
    initialize_tables,
    post_table_initialization,
    pre_table_initialization,
)


def initialize_all_database_entities(stage: Stage, components: list[Component], partner_names: list[Partner]) -> None:
    # needs default db
    pre_table_initialization(stage=stage, components=components, partner_names=partner_names)
    # needs to connect to this stage's db
    initialize_tables(stage=stage)
    # needs this stage's db
    post_table_initialization(stage=stage, components=components)


if __name__ == "__main__":
    stages = [Stage.dev, Stage.prod]
    pipeline0 = Component(
        name="pipeline0",
        grants=[
            Grant(privileges=[Privilege.SELECT, Privilege.INSERT, Privilege.UPDATE, Privilege.DELETE], tables=["event"])
        ],
        policies=[RowLevelSecurityPolicy(table="event", user_column="site_name")],
    )
    components = [pipeline0]
    partner_names = [Partner.afro_la, Partner.dallas_free_press, Partner.open_vallejo, Partner.the_19th]

    # roles are dev-pipeline0 and prod-pipeline0
    # users are dev-pipeline0-afro-la, dev-pipeline0-dallas-free-press, ... , prod-pipeline0-the-19th

    for stage in stages:
        initialize_all_database_entities(stage=stage, components=components, partner_names=partner_names)



from datetime import datetime
from enum import Enum
from typing import Optional, Union
from uuid import UUID

from pydantic import HttpUrl
from sqlmodel import JSON, Column, Field, SQLModel, String


class RefrMedium(str, Enum):
    email = "email"
    internal = "internal"
    paid = "paid"
    search = "search"
    social = "social"
    unknown = "unknown"


class Event(SQLModel, table=True):
    # TODO make enum
    site_name: str = Field(primary_key=True)
    # Timestamp making allowance for inaccurate device clock
    derived_tstamp: datetime
    # The page's height in pixels
    doc_height: float
    # Number of the current user session, e.g. first session is 1, next session is 2, etc. Dependent on domain_userid
    domain_sessionidx: int
    # User ID set by Snowplow using 1st party cookie
    domain_userid: UUID
    # Screen height in pixels. Almost 1-to-1 relationship with domain_userid (there are exceptions)
    dvce_screenheight: float
    # Screen width in pixels. Almost 1-to-1 relationship with domain_userid (there are exceptions)
    dvce_screenwidth: float
    # ID of event. This would be the primary key within the site DataFrame,
    # and part of the [site_name, event_id] composite key in the database table
    event_id: UUID = Field(primary_key=True)
    # Name of event. Can be "page_view", "page_ping", "focus_form", "change_form", "submit_form"
    # TODO make enum
    event_name: str
    # [STR, CATEGORICAL if needed] Path to page, e.g., /event-directory/ in https://dallasfreepress.com/event-directory/
    page_urlpath: str
    # URL of the referrer
    page_referrer: HttpUrl = Field(sa_column=Column(String))
    # Maximum page y-offset seen in the last ping period. Depends on event_name == "page_ping"
    pp_yoffset_max: Optional[float] = None
    # Type of referer. Can be "social", "search", "internal", "unknown", "email"
    # (read: https://docs.snowplow.io/docs/enriching-your-data/available-enrichments/referrer-parser-enrichment/)
    refr_medium: Optional[RefrMedium] = None
    # Name of referer if recognised, e.g., "Google" or "Bing"
    refr_source: Optional[str] = None
    # Data/attributes of HTML form and all its inputs in JSON format. Only present if event_name == "submit_form"
    # (read: https://github.com/snowplow/iglu-central/blob/master/schemas/com.snowplowanalytics.snowplow/submit_form/jsonschema/1-0-0)
    unstruct_event_com_snowplowanalytics_snowplow_submit_form_1: Optional[Union[list, dict]] = Field(sa_column=Column(JSON))  # type: ignore
    # Raw useragent
    useragent: str
    br_viewheight: Optional[float]
    br_viewwidth: Optional[float]
    refr_urlhost: Optional[str]
    refr_urlpath: Optional[str]


class Prescription(SQLModel, table=True):
    user_id: UUID = Field(primary_key=True)
    site_name: str = Field(primary_key=True)
    prescribe: bool
    last_updated: datetime
    # TODO will add this once we have models to work with!
    # model_id: UUID = Field(foreign_key="model.id")



from ata_db_models.helpers import (
    Component,
    Grant,
    Partner,
    Privilege,
    RowLevelSecurityPolicy,
    Stage,
    post_table_initialization,
    pre_table_initialization,
)


def add_pipeline_1(stage: Stage, components: list[Component], partner_names: list[Partner]) -> None:
    pre_table_initialization(stage=stage, components=components, partner_names=partner_names, create_dbs=False)
    # SQLAlchemy is idempotent so running this wouldn't break anything, but we don't need to run it
    # initialize_tables(stage=stage)
    post_table_initialization(stage=stage, components=components)


if __name__ == "__main__":
    stages = [Stage.dev, Stage.prod]
    pipeline1 = Component(
        name="pipeline1",
        grants=[
            Grant(privileges=[Privilege.SELECT], tables=["event"]),
            Grant(
                privileges=[Privilege.SELECT, Privilege.INSERT, Privilege.UPDATE, Privilege.DELETE],
                tables=["prescription"],
            ),
        ],
        policies=[
            RowLevelSecurityPolicy(table="event", user_column="site_name"),
            RowLevelSecurityPolicy(table="prescription", user_column="site_name"),
        ],
    )
    components = [pipeline1]
    partner_names = [Partner.afro_la, Partner.dallas_free_press, Partner.open_vallejo, Partner.the_19th]

    for stage in stages:
        add_pipeline_1(stage=stage, components=components, partner_names=partner_names)
